#!/usr/bin/perl -w

# Author: Tom Browder, <tbrowde@cox.net>, 9 Jan 2005.
# Provides for querying auditctl logs, and adding and deleting audictl
# rules.
# Free use by all allowed for any legal purpose.

use AUDIT (
	   '%syscalls',
	   '@mydirs',   # directories or file I want to monitor
	   '%mycalls',  # system calls (names) I want to monitor
	   '%myusrs',   # user names I want to monitor
	  );

my $nargs = scalar(@ARGV);
if (!$nargs) {
  print("Control auditd as user root (query, delete, or add rules):\n");
  print("Usage 1: $0 -q <audit log> [--debug]\n");
  print("  Report failures of the following system calls:\n");
  my $i = 0;
  foreach my $c (keys %mycalls) {
    if (!$i) { print("    $c"); }
    else { print(", $c"); }
    ++$i;
  }
  print("\n");
  print("  for the following users:\n");
  $i = 0;
  foreach my $u (keys %myusrs) {
    if (!$i) { print("    $u"); }
    else { print(", $u"); }
    ++$i;
  }
  print("\n");
  print("  for the following directories:\n");
  for (my $i = 0; $i < scalar(@mydirs); ++$i) {
    if (!$i) { print("    $mydirs[$i]"); }
    else { print(", $mydirs[$i]"); }
  }
  print("\n");
  print("Usage 2: $0 -D [--debug]\n");
  print("  Delete all existing rules.\n");
  print("Usage 3: $0 -A [--debug]\n");
  print("  Add rules for all users, directories, and system calls.\n");

  exit;
}
my $debug = 0;
my $log = 0;
my $add_= 0;
my $delete = 0;
my $exe = '/sbin/auditctl';
my $exefound = 0;
my $type = ''; # query, add rules, delete rules
for (my $i = 0; $i < $nargs; ++$i) {
  my $a = $ARGV[$i];
  if ($a =~ m/^-q/o) {
    if (length $a > 2) { $a =~ s/^-q//; }
    else { $a = $ARGV[++$i]; }
    $log = $a;
    $type = 'query log';
  }
  elsif ($a eq '-D') {      $delete = 1; $type = 'delete rules'; }
  elsif ($a eq '-A') {      $add    = 1; $type = 'add rules'; }
  elsif ($a eq '--debug') { $debug = 1; }
  else {
    print("ERROR: Unknown arg '$a'...exiting.");
    exit;
  }
}
if (($log && $add) || ($log && $delete) || ($add && $delete)) {
  print("ERROR: Only one type of usage at a time allowed.\n");
  exit;
}
elsif(!($log || $add || $delete)) {
  print("ERROR: No use option entered.\n");
  exit;
}
if ($log && ! -e $log) {
  print("ERROR: Unable to open file '$log'...exiting.\n");
  exit;
}
if (! -e $exe) {
  if (!$debug) {
    print("ERROR:  Binary '$exe' not available.\n");
    print("ERROR:  Use the '--debug' option to experiment further.\n");
  }
  else {
    print("NOTE:  Binary '$exe' not available.\n");
    print("       Using the '--debug' option to experiment further.\n");
  }
}
else { $exefound = 1; }

# translate system calls to ints:
my %calls = ();
foreach my $call_name (keys %mycalls) {
  if (!exists($syscalls{$call_name})) {
    die "Unknown syscall '$call_name'.\n";
  }
  my $call_number = $syscalls{$call_name};
  $calls{$call_number} = $call_name;
}

# translate user names to ints:
my %uids = ();
open(FP, "</etc/passwd");
while (defined(my $line = <FP>)) {
  # a passwd line:
  # clamav:x:501:501:Clam AntiVirus:/home/clamav:/bin/false

  my @d = split(':', $line);
  my $usr_name = $d[0];
  my $uid = $d[2];
  if (exists($myusrs{$usr_name})) {
    $myusrs{$usr_name} = $uid;
    $uids{$uid} = $usr_name;
  }
}
close(FP);

# have we found all users?
my $errors = 0;
foreach my $usr (keys %myusrs) {
  if ($myusrs{$usr} < 1) {
    print("ERROR: Unknown user '$usr'.\n");
    ++$errors;
  }
}
if ($errors && !$debug) {
  print("Exiting on usr errors.\n");
  exit;
}

# get down to work
if ($delete) { goto DELETE; }
elsif ($add) { goto ADD; }
# QUERY log =================================================================
if (!$log) { goto DELETE; }
# build a regex for failed syscalls
# e.g., $regex = '.*(syscall=10 |syscall=9 ).*(exit=[^0] )*( uid=507 | uid=2 )'
my $regex = '(';
my $i = 0;
# get all the system calls
foreach $sysnum (keys %calls) {
  if ($i) {
    $regex .= '|';
  }
  $regex .= 'syscall=' . $sysnum . ' ';
  ++$i;
}
# for now assume all errors are exit=-1
if (0) {
  $regex .= ').*(exit=-1 ).*('; # when autitd is fixed
}
else {
  $regex .= ').*(exit=[^0]* ).*('; # until auditd is fixed
}

$i = 0;
foreach $usr (keys %myusrs) {
  if ($i) {
    $regex .= '|';
  }
  $regex .= ' uid=' . $myusrs{$usr} . ' ';
  ++$i;
}
$regex .= ').*';
# testing manual regex: this works
#$regex = '(syscall=10 |syscall=9 )(exit=[^0]* ).*( uid=500 | uid=2 )';
if ($debug) {
  print("DEBUG: \$regex = '$regex'.\n");
}

open(FP, "<$log");
my $lastline = 0;
while (defined(my $line = <FP>)) {
  # matching msg pairs

  # (lastline: interested in name)
  # type=KERNEL msg=audit(1105106227.435:12373636): item=0 name=/etc/passwd \
  #   inode=5079041 dev=00:00-l inode=3080372 dev=00:00loginuid=-1 \
  #   uid=507 gid=503 euid=507 suid=507 fsuid=507 egid=503 sgid=503 fsgid=50333gid=503

  # (current line from regex match: interested in syscall, exit code, items, and uid)
  # type=KERNEL msg=audit(1105106227.435:12373636): syscall=10 exit=4294967283 \
  #   a0=bff86b22 a1=0 a2=80505e4 a3=bff86b22 items=1 pid=6137 loginuid=-1 \
  #   uid=507 gid=503 euid=507 suid=507 fsuid=507 egid=503 sgid=503 fsgid=503503
  if (0 && $debug) {
    print($line);
  }
  if ($line =~ m/$regex/o) {
    # a failure line: we may be interested also in data from the last line
    # extract and report

    # from regex line
    my ($syscall, $uid, $items, $exit); # undefined
    # from lastline
    my ($name); # undefined

    # (current line from regex match: interested in syscall, exit code, items, and uid)
    @d = split(' ', $line);
    foreach my $token (@d) {
      if ($token =~ m/^syscall=/o) {
	$syscall = $token;
	$syscall =~ s/^syscall=//o;
      }
      elsif ($token =~ m/^exit=/o) {
	$exit = $token;
	$exit =~ s/^exit=//o;
      }
      elsif ($token =~ m/^items=/o) {
	$items = $token;
	$items =~ s/^items=//o;
      }
      elsif ($token =~ m/^uid=/o) {
	$uid = $token;
	$uid =~ s/^uid=//o;
	last;
      }
    }

    # this is a hack until auditd is fixed (%d versus %u)
    # skip exit codes > 0;
    if ($exit >= 0 && $exit <  1000000) {
      $lastline = 0;
      next;
    }
    if ($items && $lastline) {
      # (lastline: interested in name)
      my @d = split(' ', $lastline);
      foreach my $token (@d) {
	if ($token =~ m/^name=/o) {
	  $name = $token;
	  $name =~ s/^name=//o;
	  last;
	}
      }
    }

    if ($debug) {
      print("DEBUG: \$syscall = '$syscall' (int).\n");
      print("DEBUG: \$syscall = '$calls{$syscall}' (text).\n");
      print("DEBUG: \$uid     = '$uid' (int).\n");
      print("DEBUG: \$uid     = '$uids{$uid}' (text).\n");
      print("DEBUG: \$items   = '$items'.\n");
      if ($items && defined($name)) {
	print("DEBUG: \$name    = '$name'.\n");
	foreach my $d (@mydirs) {
	  if ($name =~ /^$d/) {
	    print("*** NAME '$name' is of interest ***\n");
	  }
	}
      }
      else {
	print("DEBUG: \$name    = N/A.\n");
      }
      print("DEBUG: \$exit    = '$exit'.\n");
    }
    if (defined($name)) {
      my $interest = 0;
      foreach my $d (@mydirs) {
	if ($name =~ /^$d/) {
	  $interest = 0;
	  last; # one message has only one name as far as I know
	}
      }
      if ($interest) {
	  print("*** Error line and NAME '$name' is of interest ***\n");
	  print("    \$syscall = '$syscall' (int).\n");
	  print("    \$syscall = '$calls{$syscall}' (text).\n");
	  print("    \$uid     = '$uid' (int).\n");
	  print("    \$uid     = '$uids{$uid}' (text).\n");
	  print("    \$items   = '$items'.\n");
      }
    }

    # current line
    $lastline = 0;
  }
  else {
    $lastline = $line;
  }
}
close(FP);
goto FINISH;
# end QUERY log =============================================================

# DELETE rules ==============================================================
DELETE:
# delete all rules
# must execute '/sbin/auditctl -l'
if (!$exefound ) {
  print("ERROR:  Executable '$exe' not found...unable to continue with the '-D' option.\n");
}
else {
  qx( $exe -l > t );
  local(*FP);
  open(FP, "<t");
  while (defined(my $line = <FP>)) {
    # typical:
    # AUDIT_LIST: exit always uid=507 (0x1fb) exit!=0 syscall=unlink
    my @d = split(' ', $line);
    if (!defined($d[0])) { next; }
    my $f1 = shift @d;
    if ($f1 ne 'AUDIT_LIST:') { next; }
    my @F = ();
    my $call_number;
    my $arg0 = shift @d;
    my $arg1 = shift @d;
    for (my $i = 0; $i < scalar(@d); ++$i) {
      my $arg  = $d[$i];
      if (($arg =~ m/^\(/o) && ($arg =~ m/\)$/o)) {
	# skip (...) tokens
	next;
      }
      elsif ($arg =~ m/^syscall=/o) {
	$arg =~ s/^syscall=//o;
	if (!exists($syscalls{$arg})) {
	  print("ERROR: Unknown syscall '$arg'...skipping with trepidation.\n");
	  next;
	}
	$call_number = $syscalls{$arg};
      }
      elsif ($arg =~ m/=/o) {
	# assume an -F k=v pair
	push @F, $arg;
      }
    }
    # assemble the delete line
    my $cmd = "${exe} -d ${arg0},${arg1} ";
    foreach my $f (@F) {
      $cmd .= ' -F ' . $f;
    }
    $cmd .= ' -S ' . $call_number;
    print("$cmd\n");
    qx( $cmd );
  }
}
goto FINISH;
# end DELETE rules ==============================================================

# ADD rules =================================================================
ADD:
if (!$add) { goto FINISH; }
if (!$exefound ) {
  print("WARNING:  Executable '$exe' not found...unable to execute '-A' option.\n");
  print("          Debug output only...\n");
}
# Add all rules:
#foreach my $u (keys %uids) {  # prior to version 0.6.1, rules need user id
foreach my $u (keys %myusrs) { # version 0.6.1 allows rule by user name
  foreach my $c (keys %calls) {
    #  my $cmd = "$exe -a exit,always -F uid=$u -F exit!=0 -S $c\n";
    # assumes kernel audit code is working
    my $cmd = "$exe -a exit,always -F uid=$u -F success=0 -S $c\n";
    print($cmd);
    if ($exefound) {
      qx( $cmd );
    }
  }
}
# end ADD rules =================================================================

FINISH:

print("Normal end.  Type run was '$type'.\n");
